#include <stdio.h>
#include <eican.h>

// local functions
static void DisplayCardInfo( CanCardInfo *info );

/*
 * This is a very simple example program that demonstraits how to use the C++ language
 * API to receive messages over the CAN card.  The program opens the card and listens
 * for new messages.  It prints out any messages received.
 */
int main( void )
{
   // Open the card.  This opens port 0 with a bit rate of 
   // 1 Mbit/sec.  Edit the parameters to use a different 
   // bit rate or open a different port.
   EIcan cc;
   int err = cc.Open( 0, EICAN_BITRATE_1000000 );
   if( err )
   {
      printf( "Error 0x%04x opening CAN port\n", err );
      return 1;
   }

   // Read and display some information about the card
   CanCardInfo info;
   err = cc.ReadCardInfo( &info );
   if( err )
      printf( "Error 0x%04x reading information about card\n", err );
   else
      DisplayCardInfo( &info );

   // Main loop, wait for new messages and display them
   while( 1 )
   {
      printf( "Hit enter to transmit\n" );
      getchar();

      // Wait for a new message with a timeout of 5 seconds
      CanFrame frame;
      frame.type = CAN_FRAME_DATA;
      frame.id   = 0x0601;
      frame.length = 8;
      frame.data[0] = 0x40;
      frame.data[1] = 0x18;
      frame.data[2] = 0x10;
      frame.data[3] = 0x01;
      frame.data[4] = 0x00;
      frame.data[5] = 0x00;
      frame.data[6] = 0x00;
      frame.data[7] = 0x00;

      err = cc.Xmit( &frame, 500 );
      if( err )
      {
         printf( "Error transmitting: %d\n", err );
         continue;
      }

      printf( "Waiting for response\n" );
      err = cc.Recv( &frame, 5000 );
      if( err == EICAN_ERR_TIMEOUT )
      {
         printf( "No response\n" );
         continue;
      }
      else if( err )
      {
         printf( "Error 0x%04x waiting on message\n", err );
         break;
      }
      else if( frame.type == CAN_FRAME_ERROR )
      {
         CanErrorInfo *errInfo = (CanErrorInfo*)&frame;
         printf( "Error information frame received.  Error type 0x%08x, tx err count: %d, rx err count: %d\n", 
                 errInfo->mask, errInfo->data[0], errInfo->data[1] );
      }
      else if( frame.type == CAN_FRAME_REMOTE )
         printf( "Remote request, ID: 0x%08x, data count: %d", frame.id, frame.length );
      else
      {
         printf( "Received: 0x%08x ", frame.id );
         for( int i=0; i<frame.length; i++ )
            printf( "0x%02x ", frame.data[i] );
         printf( "\n" );
      }
   }

   // Since that was a while(1) loop above, we never
   // really get here.  If we did, this is how to close 
   // the port
   printf( "Closing CAN port\n" );
   err = cc.Close();
   if( err )
      printf( "Error 0x%04x closing port\n", err );

   return 0;
}

static const char *FormatVersion( uint32_t ver )
{
   static char buff[20];
   int major   = (ver & 0x00FF0000) >> 16;
   int minor   = (ver & 0x0000FF00) >> 8;
   int release = (ver & 0x000000FF);
   sprintf( buff, "%d.%02d.%02d", major, minor, release );
   return buff;
}

static void DisplayCardInfo( CanCardInfo *info )
{
   printf( "Found CAN card type: 0x%04x, serial number: %d\n", info->hwType, info->serial );
   printf( "   Firmware version: %s\n", FormatVersion( info->fwVer ) );
   printf( "Boot loader version: %s\n", FormatVersion( info->bootVer ) );
   printf( " FPGA image version: %s\n", FormatVersion( info->fpgaVer ) );
   printf( "     Driver version: %s\n", FormatVersion( info->driverVer ) );
}

